/*
 *      tcpvsadm - TCP Virtual Server ADMinistration program
 *
 *      Version: $Id: tcpvsadm.c,v 1.3 2001/10/11 01:08:12 wensong Exp $
 *
 *      Authors: Wensong Zhang <wensong@linuxvirtualserver.org>
 *
 *      Note that a lot code is taken from ipvsadm.c.
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#undef __KERNEL__       /* Makefile lazyness ;) */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <getopt.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/param.h>
#include <arpa/inet.h>

#include <net/if.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>

#include "tcp_vs.h"
#include "libtcpvs/libtcpvs.h"

static const char *program;
#define TCPVSADM_VERSION_DATE	"10-Oct-2001"
static const char *program_version = VERSION ", " TCPVSADM_VERSION_DATE;

/* default scheduler */
#define DEF_SCHED	"wlc"

/* printing format flags */
#define FMT_NONE	0x0000
#define FMT_NUMERIC	0x0001

/* various parsing helpers & parsing functions */
static int string_to_number(const char *s, int min, int max);
static int host_to_addr(const char *name, struct in_addr *addr);
static char * addr_to_host(struct in_addr *addr);
static char * addr_to_anyname(struct in_addr *addr);
static int service_to_port(const char *name, unsigned short proto);
static char * port_to_service(int port, unsigned short proto);
static char * port_to_anyname(int port, unsigned short proto);
static char * addrport_to_anyname(struct in_addr *addr, int port,
				  unsigned short proto, unsigned int format);
static int parse_service(char *buf, u_int16_t proto,
			 u_int32_t *addr, u_int16_t *port);

/* check the options based on the commands_v_options table */
static void generic_opt_check(int command, int options);
static void set_command(unsigned int *cmd, unsigned int newcmd);
static void set_option(unsigned int *options, unsigned int option);

static void tryhelp_exit(const int status);
static void usage_exit(const int status);
static void fail(int err, char *msg, ...);

static int list_service(struct tcp_vs_ident *id, unsigned int format);
static int list_all(unsigned int format);


#define CMD_NONE		0x0000U
#define CMD_ADD			0x0001U
#define CMD_EDIT		0x0002U
#define CMD_DEL			0x0004U
#define CMD_FLUSH		0x0008U
#define CMD_LIST		0x0010U
#define CMD_ADDDEST		0x0020U
#define CMD_EDITDEST		0x0040U
#define CMD_DELDEST		0x0080U
#define CMD_ADDRULE		0x0100U
#define CMD_DELRULE		0x0200U
#define CMD_START		0x0400U
#define CMD_STOP		0x0800U
#define NUMBER_OF_CMD		12

static const char* cmdnames[] = {
	"add-service",
	"edit-service",
	"delele-service",
	"flush",
	"list",
	"add-server",
	"delete-server",
	"edit-server",
	"add-rule",
	"del-rule",
	"start",
	"stop",
};

#define OPT_NONE	0x00000
#define OPT_NUMERIC	0x00001
#define OPT_IDENT	0x00002
#define OPT_SCHEDULER	0x00004
#define OPT_SERVERADDR	0x00008
#define OPT_SERVERPORT	0x00010
#define OPT_REALSERVER	0x00020
#define OPT_WEIGHT	0x00040
#define OPT_PATTERN	0x00080
#define NUMBER_OF_OPT	8

static const char* optnames[] = {
	"numeric",
	"ident",
	"scheduler",
	"serveraddr",
	"serverport",
	"real-server",
	"weight",
	"pattern",
};

/*
 * Table of legal combinations of commands and options.
 * Key:
 *  '+'  compulsory
 *  'x'  illegal
 *  '1'  exclusive (only one '1' option can be supplied)
 *  ' '  optional
 */
static const char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
{
	/*     -n   -i   -s   ads  prt  -r   -w   -p   */
/*ADD*/       {'x', '+', ' ', ' ', ' ', 'x', 'x', 'x'},
/*EDIT*/      {'x', '+', ' ', ' ', ' ', 'x', 'x', 'x'},
/*DEL*/       {'x', '+', 'x', 'x', 'x', 'x', 'x', 'x'},
/*FLUSH*/     {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
/*LIST*/      {' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x'},
/*ADD-SERVER*/{'x', '+', 'x', 'x', 'x', '+', ' ', 'x'},
/*DEL-SERVER*/{'x', '+', 'x', 'x', 'x', '+', 'x', 'x'},
/*EDIT-SRV*/  {'x', '+', 'x', 'x', 'x', '+', ' ', 'x'},
/*ADD-RULE*/  {'x', '+', 'x', 'x', 'x', '+', 'x', '+'},
/*DEL-RULE*/  {'x', '+', 'x', 'x', 'x', '+', 'x', '+'},
/*START*/     {'x', '+', 'x', 'x', 'x', 'x', 'x', 'x'},
/*STOP*/      {'x', '+', 'x', 'x', 'x', 'x', 'x', 'x'},
};

static struct option long_options[] =
{
	{"add-service", 0, 0, 'A'},
	{"edit-service", 0, 0, 'E'},
	{"delete-service", 0, 0, 'D'},
	{"flush", 0, 0, 'F'},
	{"list", 0, 0, 'L'},
	{"add-server", 0, 0, 'a'},
	{"edit-server", 0, 0, 'e'},
	{"delete-server", 0, 0, 'd'},
	{"add-rule", 0, 0, '1'},
	{"del-rule", 0, 0, '2'},
	{"start", 0, 0, '3'},
	{"stop", 0, 0, '4'},
	{"help", 0, 0, 'h'},
	{"ident", 1, 0, 'i'},
	{"port", 1, 0, 'P'},
	{"scheduler", 1, 0, 's'},
	{"real-server", 1, 0, 'r'},
	{"weight", 1, 0, 'w'},
	{"pattern", 1, 0, 'p'},
	{"numeric", 0, 0, 'n'},
	{0, 0, 0, 0}
};


int main(int argc, char **argv)
{
	const char *optstring = "AEDFaedLlhi:s:P:r:p:n";
	int c, parse;
	unsigned int command = CMD_NONE;
	unsigned int options = OPT_NONE;
	unsigned int format  = FMT_NONE;
	struct tcp_vs_ident ident;
	struct tcp_vs_config  conf;
	struct tcp_vs_dest_u dest;
	struct tcp_vs_rule_u rule;
	int result = 0;

	if(tcpvs_init())
		fail(2, "%s", tcpvs_strerror(errno));

	/* If no other arguement, list the ktcpvs service table */
	if (argc == 1){
		list_all(format);
		tcpvs_close();
		return 0;
	}

	program = argv[0];
	memset(&conf, 0, sizeof(conf));
	memset(&dest, 0, sizeof(dest));
	memset(&rule, 0, sizeof(rule));

	/* default values */
	conf.port		= htons(8080);
	conf.maxSpareServers	= 10;
	conf.minSpareServers	= 2;
	conf.startservers	= 6;
	conf.maxClients		= 2048;
	dest.weight		= 1;
	strcpy(conf.sched_name, DEF_SCHED);

	/*
	 *	Parse options
	 */
	if ((c = getopt_long(argc, argv, optstring,
			     long_options, NULL)) == EOF)
		tryhelp_exit(-1);

	switch (c) {
	case 'A':
		set_command(&command, CMD_ADD);
		break;
	case 'E':
		set_command(&command, CMD_EDIT);
		break;
	case 'D':
		set_command(&command, CMD_DEL);
		break;
	case 'a':
		set_command(&command, CMD_ADDDEST);
		break;
	case 'e':
		set_command(&command, CMD_EDITDEST);
		break;
	case 'd':
		set_command(&command, CMD_DELDEST);
		break;
	case 'F':
		set_command(&command, CMD_FLUSH);
		break;
	case '1':
		set_command(&command, CMD_ADDRULE);
		break;
	case '2':
		set_command(&command, CMD_DELRULE);
		break;
	case '3':
		set_command(&command, CMD_START);
		break;
	case '4':
		set_command(&command, CMD_STOP);
		break;
	case 'L':
	case 'l':
		set_command(&command, CMD_LIST);
		break;
	case 'h':
		usage_exit(0);
		break;
	default:
		tryhelp_exit(-1);
	}

	while ((c=getopt_long(argc, argv, optstring,
			      long_options, NULL)) != EOF) {
		switch (c) {
		case 'i':
			set_option(&options, OPT_IDENT);
			strncpy(ident.name, optarg, KTCPVS_IDENTNAME_MAXLEN);
			break;
		case 's':
			set_option(&options, OPT_SCHEDULER);
			strncpy(conf.sched_name, optarg,
				KTCPVS_SCHEDNAME_MAXLEN);
			break;
		case 'P':
			set_option(&options, OPT_SERVERPORT);
			parse = string_to_number(optarg, 1, 65534);
			if (parse == -1)
				fail(2, "illegal server port specified");
			conf.port = htons(parse);
			break;
		case 'r':
			set_option(&options, OPT_REALSERVER);
			parse = parse_service(optarg, IPPROTO_TCP,
					      &dest.addr, &dest.port);
			if (parse == 0)
				fail(2, "illegal address:port specified");
			break;
		case 'w':
			set_option(&options, OPT_WEIGHT);
			if ((dest.weight =
			     string_to_number(optarg, 0, 65535)) == -1)
				fail(2, "illegal weight specified");
			break;
		case 'p':
			set_option(&options, OPT_PATTERN);
			strncpy(rule.pattern, optarg, KTCPVS_PATTERN_MAXLEN);
			rule.len = strlen(optarg);
			break;
		case 'n':
			set_option(&options, OPT_NUMERIC);
			format |= FMT_NUMERIC;
			break;
		default:
			fail(2, "invalid option");
		}
	}

	if (optind < argc)
		fail(2, "unknown arguments found in command line");

	generic_opt_check(command, options);

	if (command == CMD_ADDRULE || command == CMD_DELRULE) {
		rule.addr = dest.addr;
		rule.port = dest.port;
	}

	switch (command) {
	case CMD_ADD:
		result = tcpvs_add_service(&ident, &conf);
		break;
	case CMD_EDIT:
		result = tcpvs_edit_service(&ident, &conf);
		break;
	case CMD_DEL:
		result = tcpvs_del_service(&ident);
		break;
	case CMD_FLUSH:
		result = tcpvs_flush();
		break;
	case CMD_LIST:
		if (options & OPT_IDENT)
			result = list_service(&ident, format);
		else
			result = list_all(format);
		break;
	case CMD_ADDDEST:
		result = tcpvs_add_dest(&ident, &dest);
		break;
	case CMD_EDITDEST:
		result = tcpvs_edit_dest(&ident, &dest);
		break;
	case CMD_DELDEST:
		result = tcpvs_add_dest(&ident, &dest);
		break;
	case CMD_ADDRULE:
		result = tcpvs_add_rule(&ident, &rule);
		break;
	case CMD_DELRULE:
		result = tcpvs_del_rule(&ident, &rule);
		break;
	case CMD_START:
		result = tcpvs_start_service(&ident);
		break;
	case CMD_STOP:
		result = tcpvs_stop_service(&ident);
		break;
	}

	if (result)
		fprintf(stderr, "%s\n", tcpvs_strerror(errno));

	tcpvs_close();
	return 0;
}


static void
print_service(struct tcp_vs_service_u *svc, unsigned int format)
{
	struct tcp_vs_get_dests *d;
	struct tcp_vs_get_rules *r;
	struct in_addr laddr;
	char *listen;
		struct in_addr daddr;
		char *dname;
	int i;

	if (!(d = tcpvs_get_dests(svc)))
		exit(1);
	if (!(r = tcpvs_get_rules(svc)))
		exit(1);

	printf("Virtual %s {\n", svc->ident.name);

	laddr.s_addr = svc->conf.addr;
	if (!(listen = addrport_to_anyname(&laddr, ntohs(svc->conf.port),
					   IPPROTO_TCP, format)))
		fail(2, "addrport_to_anyname: %s", strerror(errno));

	printf("    listen = %s\n", listen);
	printf("    scheduler = %s\n", svc->conf.sched_name);
	printf("    startservers = %d\n", svc->conf.startservers);
	printf("    maxclients = %d\n", svc->conf.maxClients);

	/* print all the destination entries */
	for (i=0; i<d->num_dests; i++) {
		struct tcp_vs_dest_u *e = &d->entrytable[i];

		daddr.s_addr = e->addr;
		if (!(dname = addrport_to_anyname(&daddr, ntohs(e->port),
						  IPPROTO_TCP, format)))
			fail(2, "addrport_to_anyname: %s", strerror(errno));
		printf("    server = %s %d\n", dname, e->weight);
		free(dname);
	}

	/* print all the rule entries */
	for (i=0; i<r->num_rules; i++) {
		struct tcp_vs_rule_u *e = &r->entrytable[i];

		daddr.s_addr = e->addr;
		if (!(dname = addrport_to_anyname(&daddr, ntohs(e->port),
						  IPPROTO_TCP, format)))
			fail(2, "addrport_to_anyname: %s", strerror(errno));
		printf("    rule = pattern %s  use server %s\n",
		       e->pattern, dname);
		free(dname);
	}
	
	printf("}\n");
	free(listen);
	free(d);
}


static int list_service(struct tcp_vs_ident *id, unsigned int format)
{
	struct tcp_vs_service_u *svc;

	if (!(svc = tcpvs_get_service(id)))
		return -1;

	print_service(svc, format);
	free(svc);
	return 0;
}


static int list_all(unsigned int format)
{
	struct tcp_vs_get_services *get;
	int i;

	printf("TCP Virtual Server version %d.%d.%d\n",
	       NVERSION(tcpvs_info.version));

	if (!(get = tcpvs_get_services()))
		return -1;

	for (i=0; i<get->num_services; i++)
		print_service(&get->entrytable[i], format);
	free(get);
	return 0;
}


static int string_to_number(const char *s, int min, int max)
{
	long number;
	char *end;

	errno = 0;
	number = strtol(s, &end, 10);
	if (*end == '\0' && end != s) {
		/* We parsed a number, let's see if we want this. */
		if (errno != ERANGE && min <= number && number <= max)
			return number;
	}
	return -1;
}


/*
 * Get IP address and port from the argument.
 * Return 0 if failed,
 *	  1 if addr read
 *        2 if addr and port read
 */
static int
parse_service(char *buf, u_int16_t proto, u_int32_t *addr, u_int16_t *port)
{
	char *pp;
	long prt;
	struct in_addr inaddr;

	pp = strchr(buf,':');
	if (pp) *pp = '\0';

	if (inet_aton(buf, &inaddr) != 0)
		*addr = inaddr.s_addr;
	else if (host_to_addr(buf, &inaddr) != -1)
		*addr = inaddr.s_addr;
	else
		return 0;

	if (pp == NULL)
		return 1;

	if ((prt=string_to_number(pp+1, 0, 65535)) != -1)
		*port = htons(prt);
	else if ((prt=service_to_port(pp+1, proto)) != -1)
		*port = htons(prt);
	else
		return 0;

	return 2;
}


static void
generic_opt_check(int command, int options)
{
	int i, j;
	int last = 0, count = 0;

	/* Check that commands are valid with options. */
	for (i = 0; i < NUMBER_OF_CMD; i++) {
		if (command & (1<<i))
			break;
	}

	for (j = 0; j < NUMBER_OF_OPT; j++) {
		if (!(options & (1<<j))) {
			if (commands_v_options[i][j] == '+')
				fail(2, "You need to supply the '%s' "
				     "option for the '%s' command",
				     optnames[j], cmdnames[i]);
		} else {
			if (commands_v_options[i][j] == 'x')
				fail(2, "Illegal '%s' option with "
				     "the '%s' command",
				     optnames[j], cmdnames[i]);
			if (commands_v_options[i][j] == '1') {
				count++;
				if (count == 1) {
					last = j;
					continue;
				}
				fail(2, "The option '%s' conflicts with the "
				     "'%s' option in the '%s' command",
				     optnames[j], optnames[last], cmdnames[i]);
			}
		}
	}
}

static inline const char *
opt2name(int option)
{
	const char **ptr;
	for (ptr = optnames; option > 1; option >>= 1, ptr++);

	return *ptr;
}

static void
set_command(unsigned int *cmd, unsigned int newcmd)
{
	if (*cmd != CMD_NONE)
		fail(2, "multiple commands specified");
	*cmd = newcmd;
}

static void
set_option(unsigned int *options, unsigned int option)
{
	if (*options & option)
		fail(2, "multiple '%s' options specified", opt2name(option));
	*options |= option;
}


static void tryhelp_exit(const int status)
{
	fprintf(stderr, "Try `%s -h' or '%s --help' for more information.\n",
		program, program);
	exit(status);
}


static void usage_exit(const int status)
{
	FILE *stream;

	if (status != 0)
		stream = stderr;
	else
		stream = stdout;

	fprintf(stream,
		"%s %s\n"
		"Usage:\n"
		"  %s -A|E -i ident [-s scheduler] [-P port]\n"
		"  %s -D -i ident\n"
		"  %s -C\n"
		"  %s -a|e -i ident -r server-address [-w weight]\n"
		"  %s -d -i ident -r server-address\n"
		"  %s --add-rule -i ident -p pattern -r server-address\n"
		"  %s --del-rule -i ident -p pattern -r server-address\n"
		"  %s -L|l [-n]\n"
		"  %s -h\n\n",
		program, program_version, program, program, program,
		program, program, program, program, program, program);

	fprintf(stream,
		"Commands:\n"
		"Either long or short options are allowed.\n"
		"  --add-service     -A        add virtual service with options\n"
		"  --edit-service    -E        edit virtual service with options\n"
		"  --delete-service  -D        delete virtual service\n"
		"  --clear           -C        clear the whole table\n"
		"  --add-server      -a        add real server with options\n"
		"  --edit-server     -e        edit real server with options\n"
		"  --delete-server   -d        delete real server\n"
		"  --add-rule                  add rule into virtual service\n"
		"  --del-rule                  del rule into virtual service\n"
		"  --list            -L|-l     list the table\n"
		"  --help            -h        display this help message\n\n"
		);

	fprintf(stream,
		"Options:\n"
		"  --ident        -i identity          service identity\n"
		"  --scheduler    -s scheduler         one of wlc|http\n"
		"                                      the default scheduler is %s.\n"
		"  --port         -p port              service port number\n"
		"  --real-server  -r server-address    server-address is host (and port)\n"
		"  --weight       -w weight            capacity of real server\n"
		"  --numeric      -n                   numeric output of addresses and ports\n",
		DEF_SCHED);

	exit(status);
}


static void fail(int err, char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
	exit(err);
}


static int host_to_addr(const char *name, struct in_addr *addr)
{
	struct hostent *host;

	if ((host = gethostbyname(name)) != NULL) {
		if (host->h_addrtype != AF_INET ||
		    host->h_length != sizeof(struct in_addr))
			return -1;
		/* warning: we just handle h_addr_list[0] here */
		memcpy(addr, host->h_addr_list[0], sizeof(struct in_addr));
		return 0;
	}
	return -1;
}


static char * addr_to_host(struct in_addr *addr)
{
	struct hostent *host;

	if ((host = gethostbyaddr((char *) addr,
				  sizeof(struct in_addr), AF_INET)) != NULL)
		return (char *) host->h_name;
	else
		return (char *) NULL;
}


static char * addr_to_anyname(struct in_addr *addr)
{
	char *name;

	if ((name = addr_to_host(addr)) != NULL)
		return name;
	else
		return inet_ntoa(*addr);
}


static int service_to_port(const char *name, unsigned short proto)
{
	struct servent *service;

	if (proto == IPPROTO_TCP
	    && (service = getservbyname(name, "tcp")) != NULL)
		return ntohs((unsigned short) service->s_port);
	else if (proto == IPPROTO_UDP
		 && (service = getservbyname(name, "udp")) != NULL)
		return ntohs((unsigned short) service->s_port);
	else
		return -1;
}


static char * port_to_service(int port, unsigned short proto)
{
	struct servent *service;

	if (proto == IPPROTO_TCP &&
	    (service = getservbyport(htons(port), "tcp")) != NULL)
		return service->s_name;
	else if (proto == IPPROTO_UDP &&
		 (service = getservbyport(htons(port), "udp")) != NULL)
		return service->s_name;
	else
		return (char *) NULL;
}


static char * port_to_anyname(int port, unsigned short proto)
{
	char *name;
	static char buf[10];

	if ((name = port_to_service(port, proto)) != NULL)
		return name;
	else {
		sprintf(buf, "%d", port);
		return buf;
	}
}


static char *
addrport_to_anyname(struct in_addr *addr, int port,
		    unsigned short proto, unsigned int format)
{
	char *buf;

	if (!(buf=malloc(60)))
		return NULL;

	if (format & FMT_NUMERIC) {
		snprintf(buf, 60, "%s:%u",
			 inet_ntoa(*addr), port);
	} else {
		snprintf(buf, 60, "%s:%s", addr_to_anyname(addr),
			 port_to_anyname(port, proto));
	}

	return buf;
}
